
/*
 * Copyright (C) 1998,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */

/*
 * This program performs surface fitting of scattered points, creating
 * a structured grid representing a surface of the form z=f(x,y).
 *
 * usage: togrid infile outfile 
 *
 * Input and output are both ascii files. The input ascii file format is:
 *
 *

    n         // number of input points
    x1 y1 z1  // x,y,z of first input point
    x2 y2 z2  // x,y,z of second input point
    ...
    xn yn zn  // last input point
 *
 * The program computes the min & max extents of the input point, and
 * constructs an output grid which spans these extents. The resolution of
 * the output grid is controlled through the two program constants,
 * OUT_USIZE and OUT_VSIZE. Change the #define statements (below) to change
 * the output grid resolution.
 *
 * The ascii format of the output file is:
 *

    2        // number of computational dimensions
    i        // integer size of output grid in U dimension
    j        // integer size of output grid in V dimension
    1        // indicates 1 value per grid point
    x1       // the x-coordinate values of the output
    x2       // grid. there are i of these.
    ...
    xi
    y1       // the y-coordinate values of the output grid.
    y2       // there are j of these.
    ...
    yj
    g11	     // the computed grid values.	
    g12
    ....
    g1i
    g21
    g22
    ...
    g2i
    ...
    gj1
    gj2
    ...
    gji
 */

 

#include <stdio.h>


#define BIGFLOAT 1.0e+20
#define SMALLFLOAT -1.0e+20

/*
 * all user modifiable parameters are in the next few #defines.
 */

/* resolution of output grid */
#define OUT_USIZE 128
#define OUT_VSIZE 128

/* search radius size as a function of output grid cell size */
#define SEARCH_RADIUS 6.0

/* inverse distance exponent */
#define EXPONENT 2.0


/* controls how many times the search radius will be doubled
   when no input points are found within the specified search
   radius. you probably shouldn't have to ever modify this. 
   presumably if you do, your input data is too sparse */
#define MAXREPEATS 8

/* end user modifiable parameters */

typedef struct
{
    float x,y;
} vert2d;

void
bivar(int npts,
      float *inx,
      float *iny,
      float *indata,
      int usize,
      int vsize,
      float *outx,
      float *outy,
      float *odata,
      float search_radius,
      float exponent,
      float undef_value)

{
    double xdist,sr,wf,t,t2,sr2,working_search_radius;
    int i,j,l,k,d,m;
    int inpoints;
    float X,dX,Y,dY,*xptr,*yptr,*f;
    float *x,*xg,*y,*yg,*z,*zg;
    float *xp,*yp;
    float xnumer,xdenom;
    int direct_hit;
    int index=0;


    sr = search_radius;
    sr2 = search_radius * search_radius;
    wf = exponent;
    
    for (k=0;k<vsize;k++)  /* loop over grid J dimension */
    {
	for (j=0;j<usize;j++)  /* loop over grid I dimension */
	{
	    xnumer = 0.;
	    xdenom = 0.;
	    direct_hit = 0;
	    l=0;

	    for (m=0;m<MAXREPEATS;m++)
	    {
		working_search_radius = sr* (float)(1<<m);
		sr2 = working_search_radius * working_search_radius;
		
		for (i=0;i<npts;i++)  /* loop on all input scatter points */
		{
	    
		    t = (inx[i]-outx[j])*(inx[i]-outx[j]) +
			(iny[i]-outy[k])*(iny[i]-outy[k]);
	    
		    if (t < 1e-6)  /* if distance == 0 */
		    {
			direct_hit = 1;
			odata[index] = indata[i];
			break;
		    }

		    if (t < sr2)  /* if distance < search radius */
		    {
			l++;
			xdist = sqrt(t);
			t2 = 1./(pow(xdist,wf));
			xnumer = xnumer + t2 * indata[i];
			xdenom = xdenom+t2;
		    }
		}
	    
		if (direct_hit != 1)
		{
		    if (l == 0)
		    {
			if (m==MAXREPEATS)
			    odata[index] = undef_value;
			/* otherwise, we'll repeat our search with a
			 larger search radius. */
		    }
		    else
		    {
			odata[index] = xnumer/xdenom;
			break;	/* no more repeats */
		    }
		}
		else
		    break;	/* no more repeats */
	    } /* loop over repeats */
	    index++;
	}
    }
}

surfaceFit(float *x,		/* input x coords */
	   float *y,		/* input y coords */
	   float *d,		/* input z value at each point */
	   int npts,		/* input: how many input points are there? */
	   vert2d *omin,	/* input: what is min extent of output grid */
	   vert2d *omax,	/* input: what is max extent of output grid */
	   int usize,		/* input: resolution of output grid in U */
	   int vsize,		/* input: resolution of output grid in V */
	   float search_radius_size, /* input: size of search radius */
	   float *newdata)	/* return: computed elevation at output grid */
{
    /*
     * note that the output grid is implicitly defined through the
     * combination of usize,vsize and omin, omax.
     */
    
    float *outx,*outy;
	
    {
	int i,j,index;
	float u,du,v,dv;
	float search_radius,exponent= EXPONENT;
	float undef_value=0.;

	void bivar(int,float *,float *,float *,int ,int,float *,float *,float *,float,float,float);
	    
	/* first, build the output grid */
	outx = (float *)malloc(sizeof(float)*usize);
	outy = (float *)malloc(sizeof(float)*vsize);

	dv = (omax->y - omin->y)/(vsize - 1);
	du = (omax->x - omin->x)/(usize - 1);
	v = omin->y;

	if (dv > du)
	    search_radius = dv*search_radius_size;
	else
	    search_radius = du*search_radius_size;

	index = 0;

	v = omin->y;
	for (j=0;j<vsize;j++,v+=dv)
	    outy[j] = v;
	
	u = omin->x;
	for (i=0;i<usize;i++,u+=du)
	    outx[i] = u; 
	    
	bivar(npts,x,y,d,usize,vsize,outx,outy,
	      newdata,search_radius,exponent,
	      undef_value);
		      

	free(outx);
	free(outy);
    }
}

	    
main(int ac,
     char *av[])
{
    FILE *fin,*fout;
    int npts,i;
    float *x,*y,*d;
    float *newdata;
    int usize,vsize;
    int method;
    float search_radius_size;
    double dx,dy,dz;
    vert2d minext,maxext;
    float *nx,*ny;
    
    /* usage: togrid infile outfile */

    fin = fopen(av[1],"r");

    fscanf(fin,"%d",&npts);
    fprintf(stderr," number of input pts is %d \n",npts);

    x = (float *)malloc(sizeof(float)*npts);
    y = (float *)malloc(sizeof(float)*npts);
    d = (float *)malloc(sizeof(float)*npts);

    minext.x = minext.y = BIGFLOAT;
    maxext.x = maxext.y = SMALLFLOAT;

    for (i=0;i<npts;i++)
    {
	fscanf(fin,"%lf %lf %lf ",&dx,&dy,&dz);
	x[i] = (float)dx;
	y[i] = (float)dy;
	d[i] = (float)dz;

	if (dx < minext.x)
	    minext.x = dx;
	if (dx > maxext.x)
	    maxext.x = dx;

	if (dy < minext.y)
	    minext.y = dy;
	if (dy > maxext.y)
	    maxext.y = dy;
    }

    fclose(fin);

    usize = OUT_USIZE;
    vsize = OUT_VSIZE;

    nx = (float *)malloc(sizeof(float)*usize);
    ny = (float *)malloc(sizeof(float)*vsize);

    nx[0] = minext.x;
    dx = (maxext.x - minext.x)/(double)(usize-1);
    
    for (i=1;i<usize;i++)
	nx[i] = nx[i-1] + dx;

    ny[0] = minext.y;
    dy = (maxext.y - minext.y)/(double)(vsize-1);
    
    for (i=1;i<vsize;i++)
	ny[i] = ny[i-1] + dy;

    newdata = (float *)malloc(sizeof(float)*usize*vsize);

    search_radius_size = SEARCH_RADIUS;	/* proportion of grid cell size. a value
				   of 1.0 == grid cell size; 2.0 means
				   two times grid cell size for search
				   radius, etc. */
    surfaceFit(x,y,d,npts,
	    &minext,&maxext,
	    usize,vsize,
	    search_radius_size,
	    newdata);

    /* write out LBL format rectilinear field */
    fout = fopen(av[2],"w");


    fprintf(fout,"2 \n%d \n%d\n 1\n",usize,vsize);

    for (i=0;i<usize;i++)
	fprintf(fout,"%f \n",nx[i]);
    fprintf(fout,"\n");

    for (i=0;i<vsize;i++)
	fprintf(fout,"%f \n",ny[i]);
    fprintf(fout,"\n");

    
    {
	int i,j,indx=0;
	
	for (j=0;j<vsize;j++)
	    for (i=0;i<usize;i++)
		fprintf(fout,"%f \n",newdata[indx++]);
    }
    fclose(fout);
}
